import shutil
import tempfile
from random import choices
from string import ascii_letters, ascii_uppercase, digits

import pytest
from faker import Faker
from faker.providers import BaseProvider
from nio.crypto import OlmAccount
from nio.store import SqliteStore

from pantalaimon.store import ClientInfo, PanStore

faker = Faker()


class Provider(BaseProvider):
    def mx_id(self):
        return "@{}:{}".format(faker.user_name(), faker.hostname())

    def device_id(self):
        return "".join(choices(ascii_uppercase, k=10))

    def access_token(self):
        return "MDA" + "".join(choices(digits + ascii_letters, k=272))

    def client(self):
        return ClientInfo(faker.mx_id(), faker.access_token())


faker.add_provider(Provider)


@pytest.fixture
def access_token():
    return faker.access_token()


@pytest.fixture
def client():
    return faker.client()


@pytest.fixture
def tempdir():
    newpath = tempfile.mkdtemp()
    yield newpath
    shutil.rmtree(newpath)


@pytest.fixture
def panstore(tempdir):
    for _ in range(10):
        store = SqliteStore(
            faker.mx_id(),
            faker.device_id(),
            tempdir,
            "",
            "pan.db"
        )
        account = OlmAccount()
        store.save_account(account)

    store = PanStore(tempdir, "pan.db")
    return store



# Copyright 2019 The Matrix.org Foundation CIC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import defaultdict
from typing import List, Optional, Tuple

import attr
from nio.store import (Accounts, DeviceKeys, DeviceTrustState, TrustState,
                       use_database)
from peewee import (SQL, DoesNotExist, ForeignKeyField, Model, SqliteDatabase,
                    TextField)


class AccessTokens(Model):
    token = TextField()
    account = ForeignKeyField(
        model=Accounts,
        primary_key=True,
        backref="access_token",
        on_delete="CASCADE"
    )


class Servers(Model):
    name = TextField()

    class Meta:
        constraints = [SQL("UNIQUE(name)")]


class ServerUsers(Model):
    user_id = TextField()
    server = ForeignKeyField(
        model=Servers,
        column_name="server_id",
        backref="users",
        on_delete="CASCADE"
    )

    class Meta:
        constraints = [SQL("UNIQUE(user_id,server_id)")]


@attr.s
class ClientInfo:
    user_id = attr.ib(type=str)
    access_token = attr.ib(type=str)


@attr.s
class OlmDevice:
    user_id = attr.ib()
    id = attr.ib()
    fp_key = attr.ib()
    sender_key = attr.ib()
    trust_state = attr.ib()


@attr.s
class PanStore:
    store_path = attr.ib(type=str)
    database_name = attr.ib(type=str, default="pan.db")
    database = attr.ib(type=SqliteDatabase, init=False)
    database_path = attr.ib(type=str, init=False)
    models = [
        Accounts,
        AccessTokens,
        Servers,
        ServerUsers,
        DeviceKeys,
        DeviceTrustState,
    ]

    def __attrs_post_init__(self):
        self.database_path = os.path.join(
            os.path.abspath(self.store_path),
            self.database_name
        )

        self.database = self._create_database()
        self.database.connect()

        with self.database.bind_ctx(self.models):
            self.database.create_tables(self.models)

    def _create_database(self):
        return SqliteDatabase(
            self.database_path,
            pragmas={
                "foreign_keys": 1,
                "secure_delete": 1,
            }
        )

    @use_database
    def _get_account(self, user_id, device_id):
        try:
            return Accounts.get(
                    Accounts.user_id == user_id,
                    Accounts.device_id == device_id,
            )
        except DoesNotExist:
            return None

    @use_database
    def save_server_user(self, server_name, user_id):
        # type: (ClientInfo) -> None
        server, _ = Servers.get_or_create(name=server_name)

        ServerUsers.replace(
            user_id=user_id,
            server=server
        ).execute()

    @use_database
    def load_all_users(self):
        users = []

        query = Accounts.select(
                Accounts.user_id,
                Accounts.device_id,
        )

        for account in query:
            users.append((account.user_id, account.device_id))

        return users

    @use_database
    def load_users(self, server_name):
        # type: () -> List[Tuple[str, str]]
        users = []

        server = Servers.get_or_none(Servers.name == server_name)

        if not server:
            return []

        server_users = []

        for u in server.users:
            server_users.append(u.user_id)

        query = Accounts.select(
                Accounts.user_id,
                Accounts.device_id,
        ).where(Accounts.user_id.in_(server_users))

        for account in query:
            users.append((account.user_id, account.device_id))

        return users

    @use_database
    def save_access_token(self, user_id, device_id, access_token):
        account = self._get_account(user_id, device_id)
        assert account

        AccessTokens.replace(
            account=account,
            token=access_token
        ).execute()

    @use_database
    def load_access_token(self, user_id, device_id):
        # type: (str, str) -> Optional[str]
        account = self._get_account(user_id, device_id)

        if not account:
            return None

        try:
            return account.access_token[0].token
        except IndexError:
            return None

    @use_database
    def load_all_devices(self):
        # type (str, str) -> Dict[str, Dict[str, DeviceStore]]
        store = dict()

        query = Accounts.select()

        for account in query:
            device_store = defaultdict(dict)

            for d in account.device_keys:

                if d.deleted:
                    continue

                try:
                    trust_state = d.trust_state[0].state
                except IndexError:
                    trust_state = TrustState.unset

                keys = {k.key_type: k.key for k in d.keys}

                device_store[d.user_id][d.device_id] = {
                    "user_id": d.user_id,
                    "device_id": d.device_id,
                    "ed25519": keys["ed25519"],
                    "curve25519": keys["curve25519"],
                    "trust_state": trust_state.name,
                    "device_display_name": d.display_name
                }

            store[account.user_id] = device_store

        return store



import pdb

from conftest import faker


class TestClass(object):
    def test_account_loading(self, panstore):
        accounts = panstore.load_all_users()
        # pdb.set_trace()
        assert len(accounts) == 10

    def test_token_saving(self, panstore, access_token):
        accounts = panstore.load_all_users()
        user_id = accounts[0][0]
        device_id = accounts[0][1]

        panstore.save_access_token(user_id, device_id, access_token)

        token = panstore.load_access_token(user_id, device_id)
        access_token == token

    def test_server_account_storing(self, panstore):
        accounts = panstore.load_all_users()

        user_id, device_id = accounts[0]
        server = faker.hostname()

        panstore.save_server_user(server, user_id)

        server2 = faker.hostname()
        user_id2, device_id2 = accounts[1]
        panstore.save_server_user(server2, user_id2)

        server_users = panstore.load_users(server)
        assert (user_id, device_id) in server_users
        assert (user_id2, device_id2) not in server_users



# Copyright 2019 The Matrix.org Foundation CIC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections import defaultdict
from pprint import pformat
from typing import Any, Dict, Optional

from aiohttp.client_exceptions import ClientConnectionError
from nio import (AsyncClient, ClientConfig, EncryptionError, KeysQueryResponse,
                 KeyVerificationEvent, KeyVerificationKey, KeyVerificationMac,
                 KeyVerificationStart, LocalProtocolError, MegolmEvent,
                 RoomEncryptedEvent, SyncResponse)
from nio.crypto import Sas
from nio.store import SqliteStore

from pantalaimon.log import logger
from pantalaimon.thread_messages import (DaemonResponse, InviteSasSignal,
                                         SasDoneSignal, ShowSasSignal,
                                         UpdateDevicesMessage)


class PanClient(AsyncClient):
    """A wrapper class around a nio AsyncClient extending its functionality."""

    def __init__(
            self,
            homeserver,
            queue=None,
            user="",
            device_id="",
            store_path="",
            config=None,
            ssl=None,
            proxy=None
    ):
        config = config or ClientConfig(store=SqliteStore, store_name="pan.db")
        super().__init__(homeserver, user, device_id, store_path, config,
                         ssl, proxy)

        self.task = None
        self.queue = queue

        self.send_semaphores = defaultdict(asyncio.Semaphore)
        self.send_decision_queues = dict()  # type: asyncio.Queue

        self.add_to_device_callback(
            self.key_verification_cb,
            KeyVerificationEvent
        )
        self.add_event_callback(
            self.undecrypted_event_cb,
            MegolmEvent
        )
        self.key_verificatins_tasks = []
        self.key_request_tasks = []

        self.add_response_callback(
            self.keys_query_cb,
            KeysQueryResponse
        )

        self.add_response_callback(
            self.sync_tasks,
            SyncResponse
        )

    @property
    def unable_to_decrypt(self):
        """Room event signaling that the message couldn't be decrypted."""
        return {
            "type": "m.room.message",
            "content": {
                "msgtype": "m.text",
                "body": ("** Unable to decrypt: The sender's device has not "
                         "sent us the keys for this message. **")
            }
        }

    async def send_message(self, message):
        """Send a thread message to the UI thread."""
        await self.queue.put(message)

    async def send_update_devcies(self):
        message = UpdateDevicesMessage()
        await self.queue.put(message)

    async def sync_tasks(self, response):
        try:
            await asyncio.gather(*self.key_verificatins_tasks)
        except LocalProtocolError as e:
            logger.info(e)

        await asyncio.gather(*self.key_request_tasks)

        self.key_verificatins_tasks = []
        self.key_request_tasks = []

    async def keys_query_cb(self, response):
        await self.send_update_devcies()

    def undecrypted_event_cb(self, room, event):
        loop = asyncio.get_event_loop()

        logger.info("Unable to decrypt event from {} via {}.".format(
            event.sender,
            event.device_id
        ))

        if event.session_id not in self.outgoing_key_requests:
            logger.info("Requesting room key for undecrypted event.")
            task = loop.create_task(self.request_room_key(event))
            self.key_request_tasks.append(task)

    def key_verification_cb(self, event):
        logger.info("Received key verification event: {}".format(event))
        loop = asyncio.get_event_loop()

        if isinstance(event, KeyVerificationStart):
            logger.info(f"{event.sender} via {event.from_device} has started "
                        f"a key verification process.")

            message = InviteSasSignal(
                self.user_id,
                event.sender,
                event.from_device,
                event.transaction_id
            )

            task = loop.create_task(
                self.queue.put(message)
            )
            self.key_verificatins_tasks.append(task)

        elif isinstance(event, KeyVerificationKey):
            sas = self.key_verifications.get(event.transaction_id, None)
            if not sas:
                return

            device = sas.other_olm_device
            emoji = sas.get_emoji()

            message = ShowSasSignal(
                self.user_id,
                device.user_id,
                device.id,
                sas.transaction_id,
                emoji
            )

            task = loop.create_task(
                self.queue.put(message)
            )
            self.key_verificatins_tasks.append(task)

        elif isinstance(event, KeyVerificationMac):
            sas = self.key_verifications.get(event.transaction_id, None)
            if not sas:
                return
            device = sas.other_olm_device

            if sas.verified:
                task = loop.create_task(self.send_message(
                    SasDoneSignal(
                        self.user_id,
                        device.user_id,
                        device.id,
                        sas.transaction_id
                    )
                ))
                self.key_verificatins_tasks.append(task)
                task = loop.create_task(self.send_update_devcies())
                self.key_verificatins_tasks.append(task)

    def start_loop(self):
        """Start a loop that runs forever and keeps on syncing with the server.

        The loop can be stopped with the stop_loop() method.
        """
        assert not self.task

        logger.info(f"Starting sync loop for {self.user_id}")

        loop = asyncio.get_event_loop()
        timeout = 30000

        sync_filter = {
            "room": {
                "state": {"lazy_load_members": True}
            }
        }

        task = loop.create_task(self.sync_forever(timeout, sync_filter))
        self.task = task
        return task

    async def start_sas(self, message, device):
        try:
            await self.start_key_verification(device)
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.ok",
                    "Successfully started the key verification request"
                    ))
        except ClientConnectionError as e:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.connection_error",
                    str(e)
                ))

    async def accept_sas(self, message):
        user_id = message.user_id
        device_id = message.device_id

        sas = self.get_active_sas(user_id, device_id)

        if not sas:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    Sas._txid_error[0],
                    Sas._txid_error[1]
                )

            )
            return

        try:
            await self.accept_key_verification(sas.transaction_id)
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.ok",
                    "Successfully accepted the key verification request"
                    ))
        except LocalProtocolError as e:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    Sas._unexpected_message_error[0],
                    str(e)
                ))
        except ClientConnectionError as e:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.connection_error",
                    str(e)
                ))

    async def cancel_sas(self, message):
        user_id = message.user_id
        device_id = message.device_id

        sas = self.get_active_sas(user_id, device_id)

        if not sas:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    Sas._txid_error[0],
                    Sas._txid_error[1]
                )

            )
            return

        try:
            await self.cancel_key_verification(sas.transaction_id)
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.ok",
                    "Successfully canceled the key verification request"
                    ))
        except ClientConnectionError as e:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.connection_error",
                    str(e)
                ))

    async def confirm_sas(self, message):
        user_id = message.user_id
        device_id = message.device_id

        sas = self.get_active_sas(user_id, device_id)

        if not sas:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    Sas._txid_error[0],
                    Sas._txid_error[1]
                )

            )
            return

        try:
            await self.confirm_short_auth_string(sas.transaction_id)
        except ClientConnectionError as e:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.connection_error",
                    str(e)
                ))

            return

        device = sas.other_olm_device

        if sas.verified:
            await self.send_update_devcies()
            await self.send_message(
                SasDoneSignal(
                    self.user_id,
                    device.user_id,
                    device.id,
                    sas.transaction_id
                )
            )
        else:
            await self.send_message(
                DaemonResponse(
                    message.message_id,
                    self.user_id,
                    "m.ok",
                    f"Waiting for {device.user_id} to confirm."
                    ))

    async def loop_stop(self):
        """Stop the client loop."""
        logger.info("Stopping the sync loop")

        if not self.task or self.task.done():
            return

        self.task.cancel()
        await self.task

    def pan_decrypt_event(
        self,
        event_dict,
        room_id=None,
        ignore_failures=True
    ):
        # type: (Dict[Any, Any], Optional[str], bool) -> (bool)
        event = RoomEncryptedEvent.parse_event(event_dict)

        if not isinstance(event, MegolmEvent):
            logger.warn("Encrypted event is not a megolm event:"
                        "\n{}".format(pformat(event_dict)))
            return False

        if not event.room_id:
            event.room_id = room_id

        try:
            decrypted_event = self.decrypt_event(event)
            logger.info("Decrypted event: {}".format(decrypted_event))

            event_dict.update(decrypted_event.source)
            event_dict["decrypted"] = True
            event_dict["verified"] = decrypted_event.verified

            return True

        except EncryptionError as error:
            logger.warn(error)

            if ignore_failures:
                event_dict.update(self.unable_to_decrypt)
            else:
                raise

            return False

    def decrypt_messages_body(self, body, ignore_failures=True):
        # type: (Dict[Any, Any], bool) -> Dict[Any, Any]
        """Go through a messages response and decrypt megolm encrypted events.

        Args:
            body (Dict[Any, Any]): The dictionary of a Sync response.

        Returns the json response with decrypted events.
        """
        if "chunk" not in body:
            return body

        logger.info("Decrypting room messages")

        for event in body["chunk"]:
            if "type" not in event:
                continue

            if event["type"] != "m.room.encrypted":
                logger.debug("Event is not encrypted: "
                             "\n{}".format(pformat(event)))
                continue

            self.pan_decrypt_event(event, ignore_failures=ignore_failures)

        return body

    def decrypt_sync_body(self, body, ignore_failures=True):
        # type: (Dict[Any, Any], bool) -> Dict[Any, Any]
        """Go through a json sync response and decrypt megolm encrypted events.

        Args:
            body (Dict[Any, Any]): The dictionary of a Sync response.

        Returns the json response with decrypted events.
        """
        logger.info("Decrypting sync")
        for room_id, room_dict in body["rooms"]["join"].items():
            try:
                if not self.rooms[room_id].encrypted:
                    logger.info("Room {} is not encrypted skipping...".format(
                        self.rooms[room_id].display_name
                    ))
                    continue
            except KeyError:
                logger.info("Unknown room {} skipping...".format(room_id))
                continue

            for event in room_dict["timeline"]["events"]:
                if "type" not in event:
                    continue

                if event["type"] != "m.room.encrypted":
                    continue

                self.pan_decrypt_event(event, room_id, ignore_failures)

        return body



# Copyright 2019 The Matrix.org Foundation CIC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from queue import Empty

import attr
import dbus
import notify2
from gi.repository import GLib
from pydbus import SessionBus
from pydbus.generic import signal

from pantalaimon.log import logger
from pantalaimon.store import PanStore
from pantalaimon.thread_messages import (AcceptSasMessage, CancelSasMessage,
                                         CancelSendingMessage,
                                         ConfirmSasMessage, DaemonResponse,
                                         DeviceBlacklistMessage,
                                         DeviceUnblacklistMessage,
                                         DeviceUnverifyMessage,
                                         DeviceVerifyMessage,
                                         ExportKeysMessage, ImportKeysMessage,
                                         InviteSasSignal, SasDoneSignal,
                                         SendAnywaysMessage, ShowSasSignal,
                                         StartSasMessage,
                                         UnverifiedDevicesSignal,
                                         UpdateDevicesMessage,
                                         UpdateUsersMessage)


class IdCounter:
    def __init__(self):
        self._message_id = 0

    @property
    def message_id(self):
        ret = self._message_id
        self._message_id += 1

        return ret


class Control:
    """
    <node>
        <interface name='org.pantalaimon1.control'>
            <method name='ListServers'>
                <arg type='a{sa(ss)}' name='servers' direction='out'/>
            </method>

            <method name='ExportKeys'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='file_path' direction='in'/>
                <arg type='s' name='passphrase' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='ImportKeys'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='file_path' direction='in'/>
                <arg type='s' name='passphrase' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='SendAnyways'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='room_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='CancelSending'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='room_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <signal name="Response">
                <arg direction="out" type="i" name="id"/>
                <arg direction="out" type="s" name="pan_user"/>
                <arg direction="out" type="a{ss}" name="message"/>
            </signal>

            <signal name="UnverifiedDevices">
                <arg direction="out" type="s" name="pan_user"/>
                <arg direction="out" type="s" name="room_id"/>
                <arg direction="out" type="s" name="room_display_name"/>
            </signal>

        </interface>
    </node>
    """

    Response = signal()
    UnverifiedDevices = signal()

    def __init__(self, queue, store, server_list, id_counter):
        self.server_list = server_list
        self.store = store
        self.queue = queue
        self.id_counter = id_counter
        self.users = defaultdict(list)
        self.update_users()

    def update_users(self):
        for server in self.server_list:
            self.users[server.name] = self.store.load_users(server.name)

    @property
    def message_id(self):
        return self.id_counter.message_id

    def ListServers(self):
        """Return the list of pan users."""
        return self.users

    def ExportKeys(self, pan_user, filepath, passphrase):
        message = ExportKeysMessage(
            self.message_id,
            pan_user,
            filepath,
            passphrase
        )
        self.queue.put(message)
        return message.message_id

    def ImportKeys(self, pan_user, filepath, passphrase):
        message = ImportKeysMessage(
            self.message_id,
            pan_user,
            filepath,
            passphrase
        )
        self.queue.put(message)
        return message.message_id

    def SendAnyways(self, pan_user, room_id):
        message = SendAnywaysMessage(
            self.message_id,
            pan_user,
            room_id
        )
        self.queue.put(message)
        return message.message_id

    def CancelSending(self, pan_user, room_id):
        message = CancelSendingMessage(
            self.message_id,
            pan_user,
            room_id
        )
        self.queue.put(message)
        return message.message_id


class Devices:
    """
    <node>
        <interface name='org.pantalaimon1.devices'>
            <method name='List'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='aa{ss}' name='devices' direction='out'/>
            </method>

            <method name='ListUserDevices'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='aa{ss}' name='devices' direction='out'/>
            </method>

            <method name='Verify'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='Unverify'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='Blacklist'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='Unblacklist'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='StartKeyVerification'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='CancelKeyVerification'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='AcceptKeyVerification'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <method name='ConfirmKeyVerification'>
                <arg type='s' name='pan_user' direction='in'/>
                <arg type='s' name='user_id' direction='in'/>
                <arg type='s' name='device_id' direction='in'/>
                <arg type='u' name='id' direction='out'/>
            </method>

            <signal name="VerificationInvite">
                <arg direction="out" type="s" name="pan_user"/>
                <arg direction="out" type="s" name="user_id"/>
                <arg direction="out" type="s" name="device_id"/>
                <arg direction="out" type="s" name="transaction_id"/>
            </signal>

            <signal name="VerificationString">
                <arg direction="out" type="s" name="pan_user"/>
                <arg direction="out" type="s" name="user_id"/>
                <arg direction="out" type="s" name="device_id"/>
                <arg direction="out" type="s" name="transaction_id"/>
                <arg direction="out" type="a(ss)" name="emoji"/>
            </signal>

            <signal name="VerificationCancel">
                <arg direction="out" type="s" name="pan_user"/>
                <arg direction="out" type="s" name="user_id"/>
                <arg direction="out" type="s" name="device_id"/>
                <arg direction="out" type="s" name="transaction_id"/>
                <arg direction="out" type="s" name="reason"/>
                <arg direction="out" type="s" name="code"/>
            </signal>

            <signal name="VerificationDone">
                <arg direction="out" type="s" name="pan_user"/>
                <arg direction="out" type="s" name="user_id"/>
                <arg direction="out" type="s" name="device_id"/>
                <arg direction="out" type="s" name="transaction_id"/>
            </signal>

        </interface>
    </node>
    """

    VerificationInvite = signal()
    VerificationCancel = signal()
    VerificationString = signal()
    VerificationDone = signal()

    def __init__(self, queue, store, id_counter):
        self.store = store
        self.device_list = None
        self.queue = queue
        self.id_counter = id_counter
        self.update_devices()

    @property
    def message_id(self):
        return self.id_counter.message_id

    def List(self, pan_user):
        device_store = self.device_list.get(pan_user, None)

        if not device_store:
            return []

        device_list = [
            device for device_list in device_store.values() for device in
            device_list.values()
        ]

        return device_list

    def ListUserDevices(self, pan_user, user_id):
        device_store = self.device_list.get(pan_user, None)

        if not device_store:
            return []

        device_list = device_store.get(user_id, None)

        if not device_list:
            return []

        return device_list.values()

    def Verify(self, pan_user, user_id, device_id):
        message = DeviceVerifyMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def Unverify(self, pan_user, user_id, device_id):
        message = DeviceUnverifyMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def Blacklist(self, pan_user, user_id, device_id):
        message = DeviceBlacklistMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def Unblacklist(self, pan_user, user_id, device_id):
        message = DeviceUnblacklistMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def StartKeyVerification(self, pan_user, user_id, device_id):
        message = StartSasMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def CancelKeyVerification(self, pan_user, user_id, device_id):
        message = CancelSasMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def ConfirmKeyVerification(self, pan_user, user_id, device_id):
        message = ConfirmSasMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def AcceptKeyVerification(self, pan_user, user_id, device_id):
        message = AcceptSasMessage(
            self.message_id,
            pan_user,
            user_id,
            device_id
        )
        self.queue.put(message)
        return message.message_id

    def update_devices(self):
        self.device_list = self.store.load_all_devices()


@attr.s
class GlibT:
    receive_queue = attr.ib()
    send_queue = attr.ib()
    data_dir = attr.ib()
    server_list = attr.ib()
    config = attr.ib()

    loop = attr.ib(init=False)
    store = attr.ib(init=False)
    users = attr.ib(init=False)
    devices = attr.ib(init=False)
    bus = attr.ib(init=False)
    control_if = attr.ib(init=False)
    device_if = attr.ib(init=False)
    notifications = attr.ib(type=bool, default=False, init=False)

    def __attrs_post_init__(self):
        self.loop = None

        self.store = PanStore(self.data_dir)
        self.users = self.store.load_all_users()

        id_counter = IdCounter()

        self.control_if = Control(self.send_queue, self.store,
                                  self.server_list, id_counter)
        self.device_if = Devices(self.send_queue, self.store, id_counter)

        self.bus = SessionBus()
        self.bus.publish("org.pantalaimon1", self.control_if, self.device_if)

    def unverified_notification(self, message):
        notificaton = notify2.Notification(
            "Unverified devices.",
            message=(f"There are unverified devices in the room "
                     f"{message.room_display_name}.")
        )
        notificaton.set_category("im")

        def send_cb(notification, action_key, user_data):
            message = user_data
            self.control_if.SendAnyways(
                message.pan_user,
                message.room_id
            )

        def cancel_cb(notification, action_key, user_data):
            message = user_data
            self.control_if.CancelSending(
                message.pan_user,
                message.room_id
            )

        if "actions" in notify2.get_server_caps():
            notificaton.add_action(
                "send",
                "Send anyways",
                send_cb,
                message
            )
            notificaton.add_action(
                "cancel",
                "Cancel sending",
                cancel_cb,
                message
            )

        notificaton.show()

    def sas_invite_notification(self, message):
        notificaton = notify2.Notification(
            "Key verification invite",
            message=(f"{message.user_id} via {message.device_id} has started "
                     f"a key verification process.")
        )
        notificaton.set_category("im")

        def accept_cb(notification, action_key, user_data):
            message = user_data
            self.device_if.AcceptKeyVerification(
                message.pan_user,
                message.user_id,
                message.device_id
            )

        def cancel_cb(notification, action_key, user_data):
            message = user_data
            self.device_if.CancelKeyVerification(
                message.pan_user,
                message.user_id,
                message.device_id,
            )

        if "actions" in notify2.get_server_caps():
            notificaton.add_action(
                "accept",
                "Accept",
                accept_cb,
                message
            )
            notificaton.add_action(
                "cancel",
                "Cancel",
                cancel_cb,
                message
            )

        notificaton.show()

    def sas_show_notification(self, message):
        emojis = [x[0] for x in message.emoji]

        emoji_str = u"   ".join(emojis)

        notificaton = notify2.Notification(
            "Short authentication string",
            message=(f"Short authentication string for the key verification of"
                     f" {message.user_id} via {message.device_id}:\n"
                     f"{emoji_str}")
        )
        notificaton.set_category("im")

        def confirm_cb(notification, action_key, user_data):
            message = user_data
            self.device_if.ConfirmKeyVerification(
                message.pan_user,
                message.user_id,
                message.device_id
            )

        def cancel_cb(notification, action_key, user_data):
            message = user_data
            self.device_if.CancelKeyVerification(
                message.pan_user,
                message.user_id,
                message.device_id,
            )

        if "actions" in notify2.get_server_caps():
            notificaton.add_action(
                "confirm",
                "Confirm",
                confirm_cb,
                message
            )
            notificaton.add_action(
                "cancel",
                "Cancel",
                cancel_cb,
                message
            )

        notificaton.show()

    def sas_done_notification(self, message):
        notificaton = notify2.Notification(
            "Device successfully verified.",
            message=(f"Device {message.device_id} of user {message.user_id} "
                     f"successfully verified.")
        )
        notificaton.set_category("im")
        notificaton.show()

    def message_callback(self):
        try:
            message = self.receive_queue.get_nowait()
        except Empty:
            return True

        logger.debug(f"UI loop received message {message}")

        if isinstance(message, UpdateDevicesMessage):
            self.device_if.update_devices()

        elif isinstance(message, UpdateUsersMessage):
            self.control_if.update_users()

        elif isinstance(message, UnverifiedDevicesSignal):
            self.control_if.UnverifiedDevices(
                message.pan_user,
                message.room_id,
                message.room_display_name
            )

            if self.notifications:
                self.unverified_notification(message)

        elif isinstance(message, InviteSasSignal):
            self.device_if.VerificationInvite(
                message.pan_user,
                message.user_id,
                message.device_id,
                message.transaction_id
            )

            if self.notifications:
                self.sas_invite_notification(message)

        elif isinstance(message, ShowSasSignal):
            self.device_if.VerificationString(
                message.pan_user,
                message.user_id,
                message.device_id,
                message.transaction_id,
                message.emoji,
            )

            if self.notifications:
                self.sas_show_notification(message)

        elif isinstance(message, SasDoneSignal):
            self.device_if.VerificationDone(
                message.pan_user,
                message.user_id,
                message.device_id,
                message.transaction_id,
            )

            if self.notifications:
                self.sas_done_notification(message)

        elif isinstance(message, DaemonResponse):
            self.control_if.Response(
                message.message_id,
                message.pan_user,
                {
                    "code": message.code,
                    "message": message.message
                }
            )

        self.receive_queue.task_done()
        return True

    def run(self):
        self.loop = GLib.MainLoop()

        if self.config.notifications:
            try:
                notify2.init("pantalaimon", mainloop=self.loop)
                self.notifications = True
            except dbus.DBusException:
                logger.error("Notifications are enabled but no notification "
                             "server could be found, disabling notifications.")
                self.notifications = False

        GLib.timeout_add(100, self.message_callback)
        self.loop.run()

    def stop(self):
        if self.loop:
            self.loop.quit()
            self.loop = None



# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

with open('README.md') as f:
    long_description = f.read()

setup(
    name="pantalaimon",
    version="0.1",
    url="https://github.com/matrix-org/pantalaimon",
    author="The Matrix.org Team",
    author_email="poljar@termina.org.uk",
    description=("A Matrix proxy daemon that adds E2E encryption "
                 "capabilities."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache License, Version 2.0",
    packages=find_packages(),
    install_requires=[
        "attrs",
        "aiohttp",
        "appdirs",
        "click",
        "keyring",
        "logbook",
        "peewee",
        "dbus-python",
        "PyGObject",
        "pydbus",
        "janus",
        "prompt_toolkit",
        "notify2",
        "typing;python_version<'3.5'",
        "matrix-nio[e2e]"
    ],
    entry_points={
        "console_scripts": ["pantalaimon=pantalaimon.main:main",
                            "panctl=pantalaimon.panctl:main"],
    },
    zip_safe=False
)
